@@ -8,6 +8,7 @@
 #
 set -e
 
+# Compat tests
 if [ "$BASH_VERSION" \< "2.01.1" ] ; then
 	echo "$0: Needs Bash >= 2.01.1" >&2
 	exit 1
@@ -21,26 +22,84 @@ if [ -z "$EMAIL" ] ; then
 		echo $"$0: Please set the environment variable EMAIL to contain your fully qualified e-mail address and try again." >&2
 		exit 1
 	fi
-	CC="`id -un`"
-	CC=`sed -ne "s/^$CC:[^:]*:[^:]*:[^:]*:\\([^,:]*\\),*:.*/\\1/p" \
-		< /etc/passwd`" <$CC@`cat /etc/mailname`>"
+	FROM="`id -un`"
+	FROM=`sed -ne "s/^$FROM:[^:]*:[^:]*:[^:]*:\\([^,:]*\\),\\?.*:.*/\\1/p" \
+		< /etc/passwd`" <$FROM@`cat /etc/mailname`>"
 else
-	CC="$EMAIL"
+	FROM="$EMAIL"
 fi
 
-FROM="$CC"
 if [ "$(type -t realpath 2> /dev/null)" = "" ]; then REALPATH=realp; else REALPATH=realpath ; fi
 
+# Wait for a keypress and put it in $KEY
+getkey()
+{
+	stty -icanon min 1 || true 2> /dev/null
+	KEY=$(dd bs=1 count=1 2> /dev/null)
+	stty icanon || true 2> /dev/null
+	KEY="${KEY:0:1}"
+	echo
+}
+
+# Usage: yesno <prompt> "yep"|"nop" (<- default)
+#	output: REPLY
+yesno()
+{
+	while true; do
+		echo -n "$1"
+
+		getkey
+
+		# if 'n'
+		if [ "$KEY" = "${YESNO:2:1}" -o "$KEY" = "${YESNO:3:1}" ]; then
+			REPLY=nop
+			return
+		fi
+
+		# if 'y'
+		if [ "$KEY" = "${YESNO:0:1}" -o "$KEY" = "${YESNO:1:1}" ]; then
+			REPLY=yep
+			return
+		fi
+
+		# if \n
+		if [ "$KEY" = "" ]; then
+			REPLY=$2
+			return
+		fi
+	done
+}
+
 realp()
 {
-	fname="$1"
+	fname=${1%/}	# strips trailing '/'
 	while [ -L "$fname" ]; do
+		oldfname="$fname"
 		fname="$(command ls -l $fname)"
 		fname="${fname#*\> }"
+		if [ "$fname" = . ] ; then
+			fname="$(dirname $oldfname)"
+		elif echo $fname | grep -vq '^/' - ; then
+			fname="$(dirname $oldfname)/$fname"
+		fi
 	done
+	pushd $(dirname $fname) > /dev/null
+	fname=$(pwd -P)/$(basename $fname)
+	popd > /dev/null
 	echo $fname
 }
 
+createfile()
+{
+	for i in $1 $1.{1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20}; do
+		if [ ! -r $i ]; then
+			tempfile --name $i && return
+		fi
+	done
+	echo $"$0: could not create temp file!" >&2
+	exit 1
+}
+
 sendhelp()
 {
 echo "bug v@VER@"
@@ -48,27 +107,30 @@ echo $"Usage: $1 [options] <packagename> [<text>]
 A script to ease the reporting of bugs.
 
 Options:
-    -c        Do  not  include  configuration  files  in  the bug
-              report.
-    -d        Debug.  Dont send the mail to the bug tracking sys
-              tem, send it to postmaster@localhost instead.
-    -f        <packagename> is a file belonging to the package
-              and bug will search for it.
+    -c          Do  not  include  configuration  files  in  the bug
+                report.
+    -d          Debug.  Don't send the mail to the bug tracking sys
+                tem, send it to the current user instead.
+    -f          <packagename> is a file belonging to the package
+                and bug will search for it.
+    -L          When used with -f prevents bug from following
+                symlinks.
     -H special_header
-              Adds  a  custom  header  to  generated mail. Can be
-              repeated to add multiple custom headers.
-    -m        Submit a \"maintonly\" bug report. Only sent  to  the
-              maintainer not to the mailing list.
-    -p        print. Do not send the bug report instead output it
-              on stdout so that it can be redirected into a file.
-    -q        Submit a quiet bug report (Just register. No e-mail
-              forwarding to anyone)
-    -s text   Set the Subject of the bug  report.  Otherwise  bug
-              will ask for a short description of the problem.
-    -x        Do not send a CC to the submitter of the bugreport
-    -z        Do not compress the configfiles by  removing  empty
-              lines and comments.
-    -h        Shows this text.
+                Adds  a  custom  header  to  generated mail. Can be
+                repeated to add multiple custom headers.
+    -m          Submit a \"maintonly\" bug report. Only sent  to  the
+                maintainer not to the mailing list.
+    -p          print. Do not send the bug report instead output it
+                on stdout so that it can be redirected into a file.
+    -q          Submit a quiet bug report (Just register. No e-mail
+                forwarding to anyone)
+    -s text     Set the Subject of the bug  report.  Otherwise  bug
+                will ask for a short description of the problem.
+    -S severity Set the severity level of the bug report.
+    -x          Do not send a CC to the submitter of the bugreport
+    -z          Do not compress the configfiles by  removing  empty
+                lines and comments.
+    -h          Shows this text.
 
 Please don't report several unrelated bugs - especially ones in
 different packages - in one message."
@@ -77,20 +139,40 @@ different packages - in one message."
 
 if [ "$1" = "--help" ]; then sendhelp ; fi
 
-while getopts "cdfhH:mpqs:vzx-" OPT; do
+if [ "$UID" = "0" ] ; then
+	echo $"Filing bug reports with system users is insecure and not supported."
+	read -p $"Who are you? Enter a non-system user: " -er U
+	echo $"Reporting bug as user $U."
+	bugrp=$($REALPATH "$0")
+	su - $U -c "$bugrp $*"
+	exit 0
+fi
+
+if grep -q "^$(id -un):" /usr/share/base-passwd/passwd.master ; then
+	echo $0: $"Filing bug reports with system users is insecure and not supported." >&2
+	exit 1
+fi
+
+while getopts "cdfhH:Lmpqs:S:vzx-" OPT; do
 	case $OPT in
 		c)	NOCONFIG=1;;
 		d)	DEBUG=1;;
 		f)	SEARCH=1;;
 		h)	sendhelp $0;;
 		H)	CUSTOM_HEADER[${#CUSTOM_HEADER[*]}]=$OPTARG;;
+		L)	NODEREFERENCESYMLINKS=1;;
 		m)	MAINT=1;;
 		p)	PRINT=1;;
 		q)	QUIET=1;;
-		s)	SUBJECT="$OPTARG";;
+		s)	SUBJECT="$OPTARG"
+			if [ -z "$SUBJECT" ]; then
+				NOSUBJECT=yep
+			fi
+			;;
+		S)	SEVERITY="$OPTARG";;
 		v)	echo "bug @VER@" >&2
 			exit 1;;
-		x)	CC="";;
+		x)	DONTCCSUBMITTER=1;;
 		z)	NOZAP=1;;
 	        -)	break;;
 	  	*)	echo $"Usage: $0 [options] <packagename> [<text>] .." >&2
@@ -101,33 +183,44 @@ shift $(($OPTIND - 1))
 
 PACKAGE="$1"
 shift || :
-if [ "$1" != "" ]; then
-	echo $"$0: Wrong number of arguments" >&2
-	exit 1
-fi
-
-if [ "$UID" = "0" ]; then
-	echo $"$0: Filing bug reports with superuser priviledges is insecure and not supported." >&2
-	exit 1
-fi
+#if [ "$1" != "" ]; then
+#	echo $"$0: Wrong number of arguments" >&2
+#	exit 1
+#fi
 
 if [ "$SEARCH" ]; then
 	if [ -z "$PACKAGE" ]; then
 		echo $"$0: Must specify file name when using -f." 2>&2
 		exit 1
 	fi
-	FILE=$($REALPATH "$PACKAGE")
+	FILE="$PACKAGE"
+	if [ ! -e "$FILE" ]; then
+		OLDIFS="$IFS"
+		IFS=":$IFS"
+		fnfound=""
+		for dir in /sbin:/usr/sbin:$PATH; do
+			if [ -e "$dir/$FILE" ]; then
+				fnfound=yep
+				FILE="$dir/$FILE"
+				break;
+			fi
+		done
+		IFS="$OLDIFS"
+		if [ -z "$fnfound" ]; then
+			echo
+			echo $"$0: File \`$FILE' not found." 2>&2
+			exit 1	
+		fi
+	fi
+	if [ ! "$NODEREFERENCESYMLINKS" ]; then
+		FILE=$($REALPATH "$FILE")
+	fi
 	test "$PRINT" || echo -n $"Searching for package containing ${FILE}..."
 	if [ "$FILE" = "" ]; then
 		echo
 		echo $"$0: Must include parameter when using -f." 2>&2
 		exit 1
 	fi
-	if [ ! -e "$FILE" ]; then
-		echo
-		echo $"$0: File \`$FILE' not found." 2>&2
-		exit 1	
-	fi
 	PACKAGE=$(grep -l ^${FILE}\$ /var/lib/dpkg/info/*.list | head -1)
 	if [ -z "$PACKAGE" ]; then
 		echo
@@ -139,20 +232,20 @@ if [ "$SEARCH" ]; then
 	test "$PRINT" || echo $"done."
 	if [ -z "$PRINT" ]; then
 		echo $"\`$FILE' found in package \`$PACKAGE'"
-		read -erp $"Submit bug for this package? [Y|n] "
-		if [ "${REPLY:0:1}" = "${YESNO:2:1}" -o "${REPLY:0:1}" = "${YESNO:3:1}" ]; then
+		yesno $"Submit bug for this package? [Y|n] " yep
+		if [ $REPLY = nop ]; then
 			exit 0
 		fi
 	fi
 fi
 
 if [ "$PACKAGE" = "" ]; then
+	echo -----------------------------------------------------------------------------
 	echo $"Please state in which package you have found a problem, or...
 type one of these bug categories:
 
 base             General bugs in the base system
-bootdisk         Bugs in the bootdisk
-rootdisk         Bugs in the rootdisk
+boot-floppies    Bugs in the installation discs
 bugs.debian.org  The bug tracking system, @bugs.debian.org
 ftp.debian.org   Problems with the FTP site
 nonus.debian.org Problems with the non-us FTP site
@@ -170,6 +263,7 @@ lists.debian.org The mailing lists debian-*@lists.debian.org."
 	fi
 fi
 
+# Checking for valid package name
 i=$((${#PACKAGE}-1))
 while [ $i -gt 0 -o $i -eq 0 ]; do
 	case ${PACKAGE:$i:1} in
@@ -177,7 +271,7 @@ while [ $i -gt 0 -o $i -eq 0 ]; do
 			echo $"$0: Package name goes first, bug report last." >&2
 			exit 1
 			;;
-		[A-Z])
+		[[:upper:]])
 			echo $"$0: Packages names are formed with lower case letters." >&2
 			exit 1
 			;;
@@ -189,7 +283,8 @@ while [ $i -gt 0 -o $i -eq 0 ]; do
 	i=$(($i - 1))
 done
 
-shift || true
+
+#shift || true
 
 if [ "$EDITOR" = "" ]; then
 	EDITOR="$VISUAL"
@@ -206,6 +301,7 @@ if [ "$EDITOR" = "" ]; then
 	fi
 fi
 
+
 checkconf()
 { 
 # Takes two parameters filename and md5sum and checks if the file was modified
@@ -239,7 +335,7 @@ pinfo()
 
 setsubject()
 {
-	if [ "$SUBJECT" = "" ]; then
+	if [ "$NOSUBJECT" != "yep" -a "$SUBJECT" = "" ]; then
 		while [ "$SUBJECT" = "" ]; do
 #			dialog --backtitle "Bug Reporting Tool - Debian/GNU Linux" \
 #				--inputbox "Please describe your problems using $PACKAGE" \
@@ -252,12 +348,67 @@ setsubject()
 	fi
 }
 
+setseverity()
+{
+	local o="cCgGiInNwWhH"
+
+	if [ -z "$SEVERITY" -a ! -z "$PRINT" ]; then
+		SEVERITY=normal
+	fi
+
+	while [ "$SEVERITY" = "" ]; do
+		echo
+		echo -----------------------------------------------------------------------------
+		echo $"Which should be the severity for this bug report? "
+		echo $"
+       0) Feature request, or maybe a bug which is very difficult to
+       fix due to major design considerations.
+
+       1) The package fails to perform correctly on some conditions, or
+       in some systems, or fails to comply current policy documents. Most
+       bugs are in this category.
+
+       2) This bug makes the package unsuitable for a stable Debian release.
+       The package should be removed from a stable distribution if this
+       bug is not fixed.
+
+       3) Dangerous bug. Makes the package in question unusable by anyone or
+       mostly so, or causes data loss, or introduces a security hole allowing
+       access to the accounts of users who use the package.
+
+       4) Critical bug. Makes unrelated software on the system (or the whole
+       system) break, or causes serious data loss, or introduces a security
+       hole on systems where you install the package."
+       		echo
+		echo -n $"Severity? [01234] "
+		getkey
+		if [ "$KEY" = "4" ]; then
+			SEVERITY=critical
+		fi
+		if [ "$KEY" = "3" ]; then
+			SEVERITY=grave
+		fi
+		if [ "$KEY" = "2" ]; then
+			SEVERITY=important
+		fi
+		if [ "$KEY" = "" -o "$KEY" = "1" ]; then
+			SEVERITY=normal
+		fi
+		if [ "$KEY" = "0" ]; then
+			SEVERITY=wishlist
+		fi
+	done
+}
+
 template()
 {
+	if [ "$SUBJECT" ]; then
+		echo "Subject: $SUBJECT"
+	fi
  	cat <<-EOF
-		Subject: $SUBJECT
 		Package: $PACKAGE
 		Version: $VERSION
+		Severity: $SEVERITY
 
 		$*
 
@@ -285,6 +436,7 @@ if [ -f /var/lib/dpkg/info/$PACKAGE.list ]; then
 	fi
 	test "$PRINT" || echo $"done."
 	setsubject
+	setseverity
 	template "$*" >$TEMP1
 
 	if [ "$DEPS" != "" ]; then
@@ -293,6 +445,8 @@ if [ -f /var/lib/dpkg/info/$PACKAGE.list ]; then
 		DEPS=`echo "$DEPS"|tr "|,\n" "  "|sed -e "s/ ([^)]*)//g" -e "s/Pre-Depends://g" -e "s/Depends://g" -e "s/  / /g" `
 		pkgs=""
 		vpkgs=""
+		
+		# See which packages are real and which are virtual
 		for i in $DEPS; do
 			if [ -f /var/lib/dpkg/info/$i.list ]; then
 				pkgs="$pkgs $i"
@@ -300,13 +454,16 @@ if [ -f /var/lib/dpkg/info/$PACKAGE.list ]; then
 				vpkgs="$vpkgs $i"
 			fi
 		done
-		test "$pkgs" && dpkg -l $pkgs | tail +6 >>$TEMP1
+		test "$pkgs" && dpkg -l $pkgs | tail +6 | sort -u >>$TEMP1
 		unset pkgs
+
+		# Now let's search for virtual packages
 		for i in $vpkgs; do
 			X=${i//+/\\\\+}
 			TEMP2=$(tempfile -pbug)
 			TEMPS="$TEMPS $TEMP2"
-			perl -n000e "print if /^Provides:.* $X/mo && !/^Status: +[^ ]+ +[^ ]+ +config-files/mo" /var/lib/dpkg/status >$TEMP2
+			perl -n000e "print if /^Provides:.* $X\b[^-+]/mo && !/^Status: +[^ ]+ +[^ ]+ +config-files/mo" \
+				/var/lib/dpkg/status >$TEMP2
 			if [ -s $TEMP2 ]; then
 				REALPACKAGE=`head -1 $TEMP2`
 				REALPACKAGE=${REALPACKAGE#Package: }
@@ -347,23 +504,23 @@ if [ -f /var/lib/dpkg/info/$PACKAGE.list ]; then
 		test "$PRINT" || echo $"done."
 	fi
 else
-	VIRT=nop
+	VERSION="N/A"
 	for i in base bootdisk rootdisk bugs.debian.org ftp.debian.org www.debian.org manual project general kernel lists.debian.org ; do
 		if [ "$PACKAGE" = "$i" ]; then
 			PSEUDO=yep
+			VERSION=$(date -u +"%Y%m%d")
 		fi
 	done
 	if [ "$PSEUDO" != "yep" ] && ! grep -q "^Package: $PACKAGE\>" /var/lib/dpkg/available; then
 		echo $"Package $PACKAGE doesn't seem to exist!"
-		echo -n $"Submit bug anyway? [y|N] "
-		read
-		if [ ! "${REPLY:0:1}" = "${YESNO:0:1}" -a ! "${REPLY:0:1}" = "${YESNO:1:1}" ]; then
+		yesno $"Submit bug anyway? [y|N] " nop
+		if [ $REPLY = nop ]; then
 			echo $"Bug report not sent"
 			exit 1
 		fi
 	fi
-	VERSION="N/A"
 	setsubject
+	setseverity
 	template "$*" >$TEMP1
 fi
 
@@ -383,27 +540,62 @@ if [ -z "$1" -a -z "$PRINT" ]; then
 		if [ "$EDNAME" = "joe" -o \
 			"$EDNAME" = "jed" -o \
 			"$EDNAME" = "emacs" -o \
+			"$EDNAME" = "ae" -o \
+			"$EDNAME" = "gnuclient" -o \
+			"$EDNAME" = "elvis" -o \
+			"$EDNAME" = "elvisnox" -o \
+			"$EDNAME" = "nvi" -o \
+			"$EDNAME" = "vim" -o \
+			"$EDNAME" = "gvim" -o \
 			"$EDNAME" = "vi" ]; then
-			EDITARGS="+5"
+			EDITARGS="+6"
 		fi
-		$EDITOR $IFJOE $EDITARGS $TEMP1 ||
-		{
-			echo $"Editor Aborted: No Bug report sent"
-			rm $TEMP1 $TEMPBACKUP
-			exit 1
-		}
 
-		if cmp $TEMP1 $TEMPBACKUP >/dev/null; then
-			echo $"No changes: No Bug report sent."
-			rm $TEMP1 $TEMPBACKUP
-			exit 1;
-		fi
-		echo -n $"Do you really want to submit this bug report? [Y|n] "
-		stty -icanon
-		KEY=$(dd bs=1 count=1 2> /dev/null)
-		stty icanon
-		echo
-		if [ "${KEY:0:1}" = "${YESNO:2:1}" -o "${KEY:0:1}" = "${YESNO:3:1}" ]; then
+		do_edit=yep
+
+		while [ "$do_edit" = yep ]; do
+
+			$EDITOR $IFJOE $EDITARGS $TEMP1 || true
+
+			if cmp $TEMP1 $TEMPBACKUP >/dev/null; then
+				echo $"Report unmodified, it won't be sent."
+				rm $TEMP1 $TEMPBACKUP
+				exit 1;
+			fi
+
+			while true; do
+				echo -n $"Do you really want to submit this bug report? [y|n|E=edit] "
+
+				getkey
+
+				# if 'n'
+				if [ "$KEY" = "${YESNO:2:1}" -o "$KEY" = "${YESNO:3:1}" ]; then
+					REPLY=nop
+					do_edit=nop
+					break
+				fi
+
+				# if 'y'
+				if [ "$KEY" = "${YESNO:0:1}" -o "$KEY" = "${YESNO:1:1}" ]; then
+					REPLY=yep
+					do_edit=nop
+					break
+				fi
+
+				# if \n
+				if [ "$KEY" = "" -o "$KEY" = $"e" \
+						-o "$KEY" = $"E" ]; then
+					REPLY=edit
+					do_edit=yep
+					break
+				fi
+			done
+		done
+
+		if [ $REPLY = nop ]; then
+			SAVEFILE=$(createfile /var/tmp/bug.dead )
+			cat "$TEMP1" >> $SAVEFILE
+			echo $"Bug report cancelled (report text left in $SAVEFILE)."
 			exit 0
 		fi
 		rm $TEMPBACKUP
@@ -412,7 +604,7 @@ if [ -z "$1" -a -z "$PRINT" ]; then
 	done
 	SUBJECT=$X
 else
-	if [ "$SUBJECT" = "" ]; then
+	if [ "$SUBJECT" = "" -a "$NOSUBJECT" != "yep" ]; then
 		SUBJECT="$PACKAGE $VERSION"
 	fi
 fi
@@ -434,7 +626,11 @@ if [ "$MAINT" ]; then
 fi
 
 if [ "$DEBUG" ]; then
-	MADDR="postmaster@localhost"
+	MADDR="$FROM"
+else
+	if [ ! "$DONTCCSUBMITTER" ]; then
+		CC="$FROM"
+	fi
 fi
 
 {
@@ -450,21 +646,32 @@ fi
 		i=$((i+1))
 	done
 	echo
-	sed <$TEMP1 -e "1d"
+	if [ -z "$1" ]; then
+		sed <$TEMP1 -e "1d"
+	else
+		echo $@
+	fi
 } |
 {
 	{
-		test -x /usr/sbin/sendmail && /usr/sbin/sendmail -t
+		test -x /usr/sbin/sendmail && /usr/sbin/sendmail -oi -t
 	} ||
-	{ 
-		tee /var/tmp/${PACKAGE}.bug | grep -q zazazaz
-		echo $"Bug report saved in /var/tmp/${PACKAGE}.bug"
+	{
+		SAVEFILE=$( createfile /var/tmp/${PACKAGE}.bug )
+		tee "$SAVEFILE" > /dev/null
+		echo
+		echo $"Error while sending bug report. Not sent."
+		echo $"Bug report saved in $SAVEFILE"
+		false
 	}
-}
+} || SAVED=true
 
 rm $TEMP1
-echo $"Bug Report submitted to $MADDR $CC
-
-If you want to submit further information about this bug please wait
+if [ "$SAVED" = "" ]; then
+	echo $"Bug Report submitted to $MADDR $CC"
+	echo $"If you want to submit further information about this bug please wait
 to receive the bug number by e-mail. Then mail any extra information to
-nnnn@bugs.debian.org, where nnnn is the bug number."
+nnnnn@bugs.debian.org, where nnnnn is the bug number."
+else
+	exit 1
+fi
